from typing import Annotated
from fastapi import FastAPI, Depends

app = FastAPI()

class Settings:
    def __init__(self):
        pass

class Repo:
    def __init__(self):
        pass

class ServiceA:
    def __init__(self, settings: Settings, repo: Repo):
        pass

class ServiceB:
    def __init__(self, repo: Repo):
        pass


def get_repo():
    return Repo()

def get_settings():
    return Settings()

def get_serviceA(repo: Annotated[Repo, Depends(get_repo)], settings: Annotated[Settings, Depends(get_settings)]):
    return ServiceA(settings, repo)

def get_serviceB(repo: Annotated[Repo, Depends(get_repo)]):
    return ServiceB(repo)

@app.get("/")
def root(settings: Annotated[Settings, Depends(get_settings)]):
    return {"message": "Hello World"}

@app.get("/a")
def a(serviceA: Annotated[ServiceA, Depends(get_serviceA)]):
    return {"message": "Hello World"}

@app.get("/b")
def b(serviceB: Annotated[ServiceB, Depends(get_serviceB)]):
    return {"message": "Hello World"}



from inspect import _empty, signature
from typing import (
    Annotated,
    Callable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    get_args,
    get_origin,
)

import fastapi
from fastapi import Depends, FastAPI
from fastapi.routing import APIRoute
from graphviz import Digraph


def get_dependencies(callable: Callable) -> List[Type]:
    """
    Get the dependencies of a callable.
    """
    sig = signature(callable)
    dependencies: List[Type] = []
    for param in sig.parameters.values():
        # Check if the parameter has an annotation (new syntax)
        if param.annotation != _empty:
            if get_origin(param.annotation) is Annotated:
                for arg in get_args(param.annotation):
                    if arg.__class__ == fastapi.params.Depends:
                        dependencies.append(arg.dependency)
            elif get_origin(param.annotation) is Depends:
                dependencies.append(param.annotation.__args__[0])

        # If not, check if the parameter has a default value (old syntax)
        else:
            if (
                param.default != _empty
                and param.default.__class__ == fastapi.params.Depends
            ):
                dependencies.append(param.default.dependency)
    return dependencies


def build_dependency_graph(app: FastAPI) -> Digraph:
    """
    Create a dependency graph of the FastAPI application as a Digraph (in DOT
    language).
    Nodes are callables and edges are dependencies.
    """
    dot = Digraph(comment="FastAPI Dependency Graph")

    visited: Set[Callable] = set()
    stack: List[Tuple[Callable, Type]] = []

    def visit(callable: Callable, _parent: Optional[Callable] = None):
        if callable in visited:
            return
        visited.add(callable)

        dependencies = get_dependencies(callable)
        for dep in dependencies:
            stack.append((callable, dep))
            visit(dep, callable)

    for route in app.routes:
        if isinstance(route, APIRoute):
            visit(route.endpoint)

    for parent, child in stack:
        dot.node(parent.__name__)
        dot.node(child.__name__)
        dot.edge(parent.__name__, child.__name__)

    return dot


def mermaid_from_dot(dot: Digraph) -> str:
    """
    Convert a Graphviz Digraph to a Mermaid diagram.
    """
    mermaid = "---\ntitle: FastAPI dependency chain\n---\n"
    mermaid += "graph TD;\n"
    for node in dot.body:
        if "label" in node:
            name = node.split("[")[0].strip()
            label = node.split("[")[1].split("]")[0]
            mermaid += f"    {name}({label})\n"
    for edge in dot.body:
        if "->" in edge:
            source, target = edge.split("->")
            source = source.strip().split()[0]
            target = target.strip().split()[0]
            mermaid += f"    {source} --> {target}\n"
    return mermaid



import argparse

from fastapi_di_viz.utils import build_dependency_graph, mermaid_from_dot

def run(args: argparse.Namespace):
    # sanity check the app argument
    if args.app.count(":") != 1:
        raise ValueError("App argument must be in the form module_name:app_name")
    module_name, app_name = args.app.split(":")

    # Check if the module and app exist
    try:
        module = __import__(module_name, fromlist=[app_name])
    except ModuleNotFoundError:
        raise ValueError(f"Module {module_name} not found")
    try:
        app = getattr(module, app_name)
    except AttributeError:
        raise ValueError(f"App {app_name} not found in module {module_name}")

    # Build the dependency graph and convert it to the desired format
    dot = build_dependency_graph(app)
    if args.format == "dot":
        print(dot.source)
    elif args.format == "mermaid":
        print(mermaid_from_dot(dot))


def main():
    argparser = argparse.ArgumentParser(
        description="""Inspect FastAPI application DI injected dependencies

Important: Install this tool in the same virtual environment that also contains
the application that it should analyze.
""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argparser.add_argument(
        "app", type=str, help="FastAPI application to inspect, e.g., myapp.main:app"
    )
    argparser.add_argument(
        "--format",
        type=str,
        choices=["mermaid", "dot"],
        default="mermaid",
        help="Output format (mermaid, dot)",
    )
    args = argparser.parse_args()
    run(args)
